/*
 * Copyright 2017-2021 Crown Copyright
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package uk.gov.gchq.gaffer.sparkaccumulo.operation.handler.graphframe;

import org.graphframes.GraphFrame;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import uk.gov.gchq.gaffer.accumulostore.AccumuloProperties;
import uk.gov.gchq.gaffer.commonutil.StreamUtil;
import uk.gov.gchq.gaffer.commonutil.TestGroups;
import uk.gov.gchq.gaffer.data.element.Edge;
import uk.gov.gchq.gaffer.data.element.Element;
import uk.gov.gchq.gaffer.data.element.Entity;
import uk.gov.gchq.gaffer.data.elementdefinition.view.View;
import uk.gov.gchq.gaffer.graph.Graph;
import uk.gov.gchq.gaffer.graph.GraphConfig;
import uk.gov.gchq.gaffer.operation.OperationChain;
import uk.gov.gchq.gaffer.operation.OperationException;
import uk.gov.gchq.gaffer.operation.impl.Map;
import uk.gov.gchq.gaffer.operation.impl.add.AddElements;
import uk.gov.gchq.gaffer.spark.SparkSessionProvider;
import uk.gov.gchq.gaffer.spark.data.generator.RowToElementGenerator;
import uk.gov.gchq.gaffer.spark.function.GraphFrameToIterableRow;
import uk.gov.gchq.gaffer.spark.operation.graphframe.GetGraphFrameOfElements;
import uk.gov.gchq.gaffer.user.User;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;

import static org.junit.jupiter.api.Assertions.assertTrue;
import static uk.gov.gchq.gaffer.data.util.ElementUtil.assertElementEquals;

public class GetGraphFrameOfElementsHandlerTest {

    private static final AccumuloProperties PROPERTIES = AccumuloProperties.loadStoreProperties(StreamUtil.storeProps(GetGraphFrameOfElementsHandlerTest.class));

    private static final int NUM_ELEMENTS = 10;

    @BeforeEach
    public void before() {
        SparkSessionProvider.getSparkSession();
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrame() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elements.json", getSimpleElements());
        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        assertElementEquals(getSimpleElements(), results);
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrameWithVertexProperty() throws OperationException {
        // Given
        final List<Element> elements = getSimpleElements();
        for (final Element element : elements) {
            element.putProperty("vertex", "value");
        }
        final Graph graph = getGraph("/schema-GraphFrame/elementsWithVertexProperty.json", elements);
        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        assertElementEquals(getSimpleElements(), results);
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrameWithNoEdges() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elements.json", getSimpleElements());
        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        assertElementEquals(getSimpleElements().stream().filter(e -> e instanceof Entity).collect(Collectors.toList()), results);
    }

    @Test
    public void shouldBehaviourInGraphFrameWithNoEntities() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elements.json", getSimpleElements());
        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        assertElementEquals(getSimpleElements().stream().filter(e -> e instanceof Edge).collect(Collectors.toList()), results);
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrameWithNoElements() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elements.json", new ArrayList<>());
        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        // When
        final GraphFrame graphFrame = graph.execute(gfOperation, new User());

        // Then
        assertTrue(graphFrame.edges().javaRDD().isEmpty());
        assertTrue(graphFrame.vertices().javaRDD().isEmpty());
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrameWithMultipleEdgesBetweenVertices() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elementsComplex.json", getElements());
        final Edge edge1 = new Edge.Builder().group(TestGroups.EDGE_2)
                .source("1")
                .dest("B")
                .directed(true)
                .property("columnQualifier", 1)
                .property("property1", 2)
                .property("property2", 3.0F)
                .property("property3", 4.0D)
                .property("property4", 5L)
                .property("count", 100L)
                .build();

        final Edge edge2 = new Edge.Builder().group(TestGroups.EDGE_2)
                .source("1")
                .dest("C")
                .directed(true)
                .property("columnQualifier", 6)
                .property("property1", 7)
                .property("property2", 8.0F)
                .property("property3", 9.0D)
                .property("property4", 10L)
                .property("count", 200L)
                .build();

        graph.execute(new AddElements.Builder().input(edge1, edge2).build(), new User());

        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        final List<Element> expected = getElements();
        expected.add(edge1);
        expected.add(edge2);

        assertElementEquals(expected, results);
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrameWithLoops() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elementsComplex.json", getElements());
        final Edge edge1 = new Edge.Builder().group(TestGroups.EDGE)
                .source("B")
                .dest("1")
                .directed(true)
                .property("columnQualifier", 1)
                .property("property1", 2)
                .property("property2", 3.0F)
                .property("property3", 4.0D)
                .property("property4", 5L)
                .property("count", 100L)
                .build();

        final Edge edge2 = new Edge.Builder().group(TestGroups.EDGE)
                .source("C")
                .dest("1")
                .directed(true)
                .property("columnQualifier", 6)
                .property("property1", 7)
                .property("property2", 8.0F)
                .property("property3", 9.0D)
                .property("property4", 10L)
                .property("count", 200L)
                .build();

        graph.execute(new AddElements.Builder().input(edge1, edge2).build(), new User());

        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        final List<Element> expected = getElements();
        expected.add(edge1);
        expected.add(edge2);
        assertElementEquals(expected, results);
    }

    @Test
    public void shouldGetCorrectElementsInGraphFrameWithRepeatedElements() throws OperationException {
        // Given
        final Graph graph = getGraph("/schema-GraphFrame/elements.json", getSimpleElements());
        graph.execute(new AddElements.Builder().input(getSimpleElements()).build(), new User());

        final GetGraphFrameOfElements gfOperation = new GetGraphFrameOfElements.Builder()
                .view(new View.Builder()
                        .edge(TestGroups.EDGE)
                        .edge(TestGroups.EDGE_2)
                        .entity(TestGroups.ENTITY)
                        .entity(TestGroups.ENTITY_2)
                        .build())
                .build();

        final OperationChain<Iterable<? extends Element>> opChain = new OperationChain.Builder()
                .first(gfOperation)
                .then(new Map.Builder<>()
                        .first(new GraphFrameToIterableRow())
                        .then(new RowToElementGenerator())
                        .build())
                .build();

        // When
        final Iterable<? extends Element> results = graph.execute(opChain, new User());

        // Then
        assertElementEquals(getSimpleElements(), results);
    }

    private Graph getGraph(final String elementsSchema, final List<Element> elements) throws OperationException {
        final Graph graph = new Graph.Builder()
                .config(new GraphConfig.Builder()
                        .graphId("graphId")
                        .build())
                .addSchema(getClass().getResourceAsStream(elementsSchema))
                .addSchema(getClass().getResourceAsStream("/schema-GraphFrame/types.json"))
                .storeProperties(PROPERTIES)
                .build();
        graph.execute(new AddElements.Builder().input(elements).build(), new User());
        return graph;
    }

    private List<Element> getElements() {
        final List<Element> elements = new ArrayList<>();

        final Entity entityB = new Entity.Builder().group(TestGroups.ENTITY)
                .vertex("B")
                .property("columnQualifier", 1)
                .property("count", 1L)
                .build();

        final Entity entityC = new Entity.Builder().group(TestGroups.ENTITY)
                .vertex("C")
                .property("columnQualifier", 1)
                .property("count", 1L)
                .build();

        elements.add(entityB);
        elements.add(entityC);

        for (int i = 0; i < NUM_ELEMENTS; i++) {
            final Entity entity1 = new Entity.Builder().group(TestGroups.ENTITY)
                    .vertex("" + i)
                    .property("columnQualifier", 1)
                    .property("property1", i)
                    .property("property2", 3.0F)
                    .property("property3", 4.0D)
                    .property("property4", 5L)
                    .property("count", 6L)
                    .build();

            final Entity entity2 = new Entity.Builder().group(TestGroups.ENTITY_2)
                    .vertex("" + i)
                    .property("columnQualifier", 1)
                    .property("property1", i)
                    .property("property2", 3.0F)
                    .property("property3", 4.0D)
                    .property("property4", 5L)
                    .property("count", 6L)
                    .build();


            final Edge edge1 = new Edge.Builder().group(TestGroups.EDGE)
                    .source("" + i)
                    .dest("B")
                    .directed(true)
                    .property("columnQualifier", 1)
                    .property("property1", 2)
                    .property("property2", 3.0F)
                    .property("property3", 4.0D)
                    .property("property4", 5L)
                    .property("count", 100L)
                    .build();

            final Edge edge2 = new Edge.Builder().group(TestGroups.EDGE)
                    .source("" + i)
                    .dest("C")
                    .directed(true)
                    .property("columnQualifier", 6)
                    .property("property1", 7)
                    .property("property2", 8.0F)
                    .property("property3", 9.0D)
                    .property("property4", 10L)
                    .property("count", i * 200L)
                    .build();

            elements.add(edge1);
            elements.add(edge2);
            elements.add(entity1);
            elements.add(entity2);
        }
        return elements;
    }

    private List<Element> getSimpleElements() {
        final List<Element> elements = new ArrayList<>();

        elements.add(new Entity.Builder().group(TestGroups.ENTITY)
                .vertex("B")
                .property("fullname", "")
                .build());

        elements.add(new Entity.Builder().group(TestGroups.ENTITY_2)
                .vertex("C")
                .build());

        for (int i = 0; i < NUM_ELEMENTS; i++) {
            elements.add(new Entity.Builder().group(TestGroups.ENTITY)
                    .vertex("" + i)
                    .property("fullname", "name_" + i)
                    .build());

            elements.add(new Entity.Builder().group(TestGroups.ENTITY_2)
                    .vertex("" + i)
                    .build());

            elements.add(new Edge.Builder().group(TestGroups.EDGE)
                    .source("" + i)
                    .dest("B")
                    .directed(true)
                    .property("type", "type_" + i)
                    .build());

            elements.add(new Edge.Builder().group(TestGroups.EDGE)
                    .source("" + i)
                    .dest("C")
                    .directed(true)
                    .property("type", "type_" + i)
                    .build());
        }
        return elements;
    }
}
